from sidecar.aws_session import AwsSession
from sidecar.aws_tag_helper import AwsTagHelper
from sidecar.const import Const


class AwsVpcService():
    def __init__(self, sandbox_id: str, aws_session: AwsSession) -> None:
        super().__init__()
        self.aws_session = aws_session
        self.sandbox_id = sandbox_id
        self._ec2_client = self.aws_session.get_ec2_client()
        self._ec2_resource = self.aws_session.get_ec2_resource()

    def update_tags_on_vpc(self, tags: {}):
        tags_list = [AwsTagHelper.create_tag(tag_name, tag_value) for tag_name, tag_value in tags.items()]
        vpc = self._get_vpc()
        self._ec2_client.create_tags(
            Resources=[vpc.id],
            Tags=tags_list
        )

    def _get_vpc(self):
        filters = [{'Name': 'tag:' + Const.SANDBOX_ID_TAG,
                    'Values': [self.sandbox_id]}]

        vpcs = list(self._ec2_resource.vpcs.filter(Filters=filters))

        if not vpcs:
            return None

        if len(vpcs) > 1:
            raise ValueError('Too many vpcs for the sandbox')

        return vpcs[0]



from sidecar.const import Const


class AwsTagHelper:
    @staticmethod
    def get_tags(resource) -> {}:
        tags = resource.tags
        return dict((tag['Key'], tag['Value']) for tag in tags)

    @staticmethod
    def safely_get_tag(resource, tag_name: str, value_if_missing: str = None) -> str:
        return AwsTagHelper.get_tags(resource=resource).get(tag_name, value_if_missing)

    @staticmethod
    def create_tag(key, value):
        return {'Key': key, 'Value': value}

    @staticmethod
    def get_apps_status_map(instance) -> {}:
        apps_status_tag_value = AwsTagHelper.safely_get_tag(instance, Const.APP_STATUS_TAG)
        return AwsTagHelper.parse_apps_status_tag_value(apps_status_tag_value)

    @staticmethod
    def parse_apps_status_tag_value(apps_status_value: str) -> {}:
        app_to_status_map = dict()
        if apps_status_value:
            all_statuses = apps_status_value.split(Const.CSV_TAG_VALUE_SEPARATE)
            app_to_status_map = dict(state.split(Const.APP_STATE_KEY_VALUE_SEPARATOR) for state in all_statuses)
        return app_to_status_map

    @staticmethod
    def format_apps_status_tag_value(app_to_status_map: {}) -> str:
        return Const.CSV_TAG_VALUE_SEPARATE.join(
            ["{KEY}{SEP}{VALUE}".format(KEY=app_name, SEP=Const.APP_STATE_KEY_VALUE_SEPARATOR, VALUE=app_status)
             for app_name, app_status in app_to_status_map.items()])



import json
import logging
import os
from typing import List

from flask import Flask, request
from flask import jsonify

from sidecar.app_configuration_start_policy import AppConfigurationStartPolicy
from sidecar.app_request_info import AppRequestInfo
from sidecar.app_status_maintainer import AppStatusMaintainer
from sidecar.apps_configuration_end_tracker import AppsConfigurationEndTracker
from sidecar.aws_app_instance_identifier_creator import AwsAppInstanceIdentifierCreator
from sidecar.aws_app_instance_service import AwsAppInstanceService
from sidecar.aws_sandbox_start_time_updater import AwsSandboxStartTimeUpdater
from sidecar.aws_sandbox_deployment_end_updater import AwsSandboxDeploymentEndUpdater
from sidecar.aws_session import AwsSession
from sidecar.aws_vpc_service import AwsVpcService
from sidecar.const import Const, DateTimeProvider
from sidecar.file_system import FileSystemService
from sidecar.health_check_monitor import HealthCheckMonitor
from sidecar.kub_api_service import KubApiService
from sidecar.kub_app_instance_identifier_creator import KubAppInstanceIdentifierCreator
from sidecar.kub_app_instance_service import KubAppInstanceService
from sidecar.kub_sandbox_start_time_updater import KubSandboxStartTimeUpdater
from sidecar.kub_sandbox_deployment_end_updater import KubSandboxDeploymentEndUpdater
from sidecar.kub_token_provider import KubTokenProvider
from sidecar.sandbox_deployment_state_tracker import SandboxDeploymentStateTracker
from sidecar.utils import Utils

app = Flask(__name__)
logger = None
_health_check_monitor = None
_app_configuration_start_policy = None
_app_instance_identifier_creator = None


def _convert_apps_json_to_app_requests(apps_json: dict()) -> List[AppRequestInfo]:
    # TODO: remove the conditional assignment once all branches are up-to-date with issue #95
    app_requests = [AppRequestInfo(name=app_name,
                                   instances_count=app_value["instances_count"] if isinstance(app_value, dict) else app_value,
                                   app_dependencies=app_value["dependencies"] if isinstance(app_value, dict) else [])
                    for app_name, app_value in apps_json.items()]
    return app_requests


def get_status_updater():
    try:
        with open(Const.get_config_file(), 'r') as conf:
            config_json = json.loads(conf.read())

            provider = config_json["provider"]
            sandbox_id = config_json['sandbox_id']
            apps_json = config_json['apps']
            app_requests = _convert_apps_json_to_app_requests(apps_json=apps_json)
            if provider == "kubernetes":
                kub_token_provider = KubTokenProvider()
                hostname = config_json['kub_api_address']
                kas = KubApiService(hostname=hostname, namespace=sandbox_id, kub_token_provider=kub_token_provider)

                app_instance_service = KubAppInstanceService(logger=logger, kub_api_service=kas)
                apps_configuration_end_tracker = AppsConfigurationEndTracker(logger=logger, app_requests=app_requests,
                                                                             app_instance_service=app_instance_service)
                app_configuration_start_policy = AppConfigurationStartPolicy(apps_configuration_end_tracker,
                                                                             app_requests)

                sandbox_start_time_updater = KubSandboxStartTimeUpdater(
                    date_time_provider=DateTimeProvider(),
                    logger=logger,
                    kub_api_service=kas,
                    apps_configuration_end_tracker=apps_configuration_end_tracker)

                sandbox_deployment_end_updater = KubSandboxDeploymentEndUpdater(
                    kub_api_service=kas)

                sandbox_deployment_state_tracker = SandboxDeploymentStateTracker(
                    logger=logger,
                    app_requests=app_requests,
                    apps_configuration_end_tracker=apps_configuration_end_tracker,
                    sandbox_deployment_end_updater=sandbox_deployment_end_updater)

                app_instance_identifier_creator = KubAppInstanceIdentifierCreator(kub_api_service=kas)
                status_maintainer = AppStatusMaintainer(logger=logger,
                                                        app_instance_service=app_instance_service,
                                                        apps_configuration_end_tracker=apps_configuration_end_tracker,
                                                        sandbox_deployment_state_tracker=sandbox_deployment_state_tracker,
                                                        sandbox_start_time_updater=sandbox_start_time_updater)
                return status_maintainer, \
                       app_configuration_start_policy, \
                       app_instance_identifier_creator

            if provider == "aws":
                region_name = config_json['region_name']
                aws_session = AwsSession(region_name=region_name)

                app_instance_service = AwsAppInstanceService(sandbox_id=sandbox_id, logger=logger,
                                                             aws_session=aws_session)
                apps_configuration_end_tracker = AppsConfigurationEndTracker(logger=logger, app_requests=app_requests,
                                                                             app_instance_service=app_instance_service)
                app_configuration_start_policy = AppConfigurationStartPolicy(apps_configuration_end_tracker,
                                                                             app_requests)
                aws_vpc_service = AwsVpcService(sandbox_id=sandbox_id, aws_session=aws_session)
                sandbox_start_time_updater = AwsSandboxStartTimeUpdater(
                    sandbox_id=sandbox_id,
                    aws_session=aws_session,
                    date_time_provider=DateTimeProvider(),
                    logger=logger,
                    vpc_service=aws_vpc_service,
                    apps_configuration_end_tracker=apps_configuration_end_tracker)
                sandbox_deployment_end_updater = AwsSandboxDeploymentEndUpdater(vpc_service=aws_vpc_service)

                sandbox_deployment_state_tracker = SandboxDeploymentStateTracker(
                    logger=logger,
                    app_requests=app_requests,
                    apps_configuration_end_tracker=apps_configuration_end_tracker,
                    sandbox_deployment_end_updater=sandbox_deployment_end_updater)

                app_instance_identifier_creator = AwsAppInstanceIdentifierCreator()
                status_maintainer = AppStatusMaintainer(logger=logger,
                                                        app_instance_service=app_instance_service,
                                                        apps_configuration_end_tracker=apps_configuration_end_tracker,
                                                        sandbox_deployment_state_tracker=sandbox_deployment_state_tracker,
                                                        sandbox_start_time_updater=sandbox_start_time_updater)
                return status_maintainer, \
                       app_configuration_start_policy, \
                       app_instance_identifier_creator

            raise Exception('unknown provider {}'.format(provider))
    except Exception as exc:
        logger.exception(exc)
        raise exc


@app.route("/")
def hello():
    logger.info("start")
    return "welcome cs18:sidecar-api"


@app.route("/application/<string:app_name>/health-check", methods=['POST'])
def start_health_check(app_name: str):
    logger.info("entered")
    # noinspection PyBroadException
    try:
        ip_address = request.remote_addr
        req = request.get_json()
        script_name = req['script']
        additional_data = req.get('additional_data')

        logger.info("start_health_check for app '{}', address: {}, request: {}".format(app_name, ip_address, req))
        app_instance_identifier = _app_instance_identifier_creator.create(app_name=app_name,
                                                                          ip_address=ip_address,
                                                                          additional_data=additional_data)
        logger.info("created identifier for '{APP}' on '{IP}': '{APP_ID}'"
                    .format(APP=app_name, IP=ip_address, APP_ID=app_instance_identifier))
        _health_check_monitor.start(ip_address=ip_address, app_instance_identifier=app_instance_identifier,
                                    script_name=script_name)

        # fail here if validation fails ONLY
        return jsonify(success=True), 202
    except Exception as exc:
        logger.exception(exc)
        return jsonify(success=False, error=str(exc)), 401


@app.route("/application/<string:app_name>/log", methods=['GET'])
def get_application_log(app_name: str):
    logger.info("entered")
    # noinspection PyBroadException
    try:
        log = Utils.read_log(app_name=app_name)
        return jsonify(success=True, log=log), 200
    except FileNotFoundError as exc:
        logger.exception(exc)
        return jsonify(success=False, error="log for app {} was not found".format(app_name)), 400
    except Exception as exc:
        logger.exception(exc)
        return jsonify(success=False, error=str(exc)), 401


@app.route("/application/<string:app_name>/config-start-status", methods=['GET'])
def get_application_configuration_start_status(app_name: str):
    logger.info("entered")
    try:
        start_status = _app_configuration_start_policy.get_app_configuration_start_status(app_name)
        return start_status, 200
    except Exception as exc:
        logger.exception(exc)
        return str(exc), 500


# @app.route("/application/<string:app_name>/config-end-status", methods=['POST'])
# def set_application_configuration_end_status(app_name: str):
#     logger.info("entered")
#     try:
#         ip_address = request.remote_addr
#         data_json = request.get_json()
#         status = data_json['status']
#         additional_data = data_json.get('additional_data')
#
#         if not AppInstanceConfigStatus.is_end_status(status):
#             error_message = "'{STATUS}' is not one of the allowed configuration end statuses".format(STATUS=status)
#             return jsonify(success=False, error=error_message), 400
#         app_instance_identifier = _app_instance_identifier_creator.create(app_name=app_name,
#                                                                                     ip_address=ip_address,
#                                                                                     additional_data=additional_data)
#         _status_maintainer.update_status(app_instance_identifier=app_instance_identifier,
#                                          status=status)
#
#         return jsonify(success=True), 200
#     except Exception as exc:
#         logger.exception(exc)
#         return jsonify(success=False, error=str(exc)), 500


def run():
    global _health_check_monitor
    global logger
    global _app_configuration_start_policy
    global _app_instance_identifier_creator
    # logger definitions
    logger = logging.getLogger('sidecar-api')
    hdlr = logging.FileHandler(Const.get_log_file())
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(filename)s %(funcName)s -> %(message)s')
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)
    logger.setLevel(logging.INFO)

    status_maintainer, _app_configuration_start_policy, _app_instance_identifier_creator = get_status_updater()
    _health_check_monitor = HealthCheckMonitor(logger=logger,
                                               file_service=FileSystemService(),
                                               status_maintainer=status_maintainer)
    # server startstatus_updater
    app.run(host='0.0.0.0', port=4000)


def configure_debug():
    # create log folder: ~/sidecar
    api_log_folder = os.path.dirname(Const.get_log_file())
    file_system = FileSystemService()
    if not file_system.exists(api_log_folder):
        file_system.create_folders(api_log_folder)

    # setup the config file
    apps_json = {
        "myApp": 1
    }
    # apps_json = {
    #     "myApp": {
    #         "instances_count": 1,
    #         "dependencies": ["app2"]
    #     },
    #     "app2":{
    #         "instances_count": 1,
    #         "dependencies": []
    #     }
    # }
    config_file_content = {
        "provider": "aws",
        "sandbox_id": "17011808-85AD-4D83-9AAB-4E21442BFF97",  # can set it to an existing sandbox
        # "kub_api_address": "https://35.197.122.51",
        "region_name": "eu-west-2",
        "apps": apps_json
    }
    with open(Const.get_config_file(), "w") as conf_file:
        conf_file.write(json.dumps(config_file_content))


if __name__ == "__main__":
    # configure_debug()
    run()



from sidecar.app_instance_identifier import AppInstanceIdentifier
from sidecar.app_instance_identifier_creator import IAppInstanceIdentifierCreator


class AwsAppInstanceIdentifierCreator(IAppInstanceIdentifierCreator):
    def __init__(self) -> None:
        super().__init__()

    def create(self, app_name: str, ip_address: str, additional_data: {}) -> AppInstanceIdentifier:
        return AppInstanceIdentifier(name=app_name, infra_id=additional_data['instance-id'])



from sidecar.aws_vpc_service import AwsVpcService
from sidecar.const import Const
from sidecar.sandbox_deployment_end_updater import ISandboxDeploymentEndUpdater


class AwsSandboxDeploymentEndUpdater(ISandboxDeploymentEndUpdater):
    def __init__(self, vpc_service: AwsVpcService):
        super(AwsSandboxDeploymentEndUpdater, self).__init__()
        self._vpc_service = vpc_service

    def _set_deployment_end_details(self, deployment_end_status: str):
        self._vpc_service.update_tags_on_vpc({Const.SANDBOX_DEPLOYMENT_END_STATUS: deployment_end_status})



import os
import sys

import time

import datetime

from sidecar.const import Const


class Utils:
    @staticmethod
    def stop_on_debug():
        while not sys.gettrace():
            time.sleep(0.5)

    @staticmethod
    def read_log(app_name: str) -> str:
        file_path = Const.get_app_log_file(app_name)
        with open(file_path, 'r') as application_log:
            return application_log.read()

    @staticmethod
    def get_timestamp():
        return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")



import boto3


class AwsSession:
    def __init__(self, region_name: str):
        self._region_name = region_name
        self.ec2_client = self._get_client('ec2')
        self.cf_client = self._get_client('cloudformation')
        self.ec2_resource = self._get_resource('ec2')
        self._autoscalling_client = self._get_client('autoscaling')

    def get_ec2_client(self):
        return self.ec2_client

    def get_cf_client(self):
        return self.cf_client

    def get_ec2_resource(self):
        return self.ec2_resource

    def get_autoscaling_client(self):
        return self._autoscalling_client

    def _get_client(self, client: str):
        return boto3.client(client, region_name=self._region_name)

    def _get_resource(self, resource: str):
        return boto3.resource(resource, region_name=self._region_name)


import os
import signal
import _thread
from logging import Logger
import subprocess

from sidecar.app_instance_identifier import AppInstanceIdentifier
from sidecar.app_status_maintainer import AppStatusMaintainer
from sidecar.const import Const, AppInstanceConfigStatus
from sidecar.file_system import FileSystemService
from sidecar.utils import Utils


class HealthCheckMonitor:
    def __init__(self, logger: Logger, file_service: FileSystemService,
                 status_maintainer: AppStatusMaintainer):
        self._file_service = file_service
        self.logger = logger
        self._status_maintainer = status_maintainer
        self.timeout = Const.HEALTH_CHECK_TIMEOUT

    def start(self, ip_address: str, app_instance_identifier: AppInstanceIdentifier, script_name: str):
        self.logger.info("entered '{}'".format(app_instance_identifier.name))
        _thread.start_new_thread(self._process_health_check, (ip_address, app_instance_identifier, script_name))

    def _update_status(self, app_instance_identifier: AppInstanceIdentifier, status: str):
        self._status_maintainer.update_status(app_instance_identifier=app_instance_identifier,
                                              status=status)

    def _process_health_check(self, ip_address: str, app_instance_identifier: AppInstanceIdentifier, script_name: str):
        try:
            self._update_status(app_instance_identifier=app_instance_identifier,
                                status=AppInstanceConfigStatus.PENDING)

            result = self._execute_health_check_script(app_name=app_instance_identifier.name,
                                                       ip_address=ip_address,
                                                       script_name=script_name)

            status = AppInstanceConfigStatus.COMPLETED if result else AppInstanceConfigStatus.ERROR
            self._update_status(app_instance_identifier=app_instance_identifier,
                                status=status)
        except Exception as exc:
            self.logger.exception(exc)
            raise

    def _execute_health_check_script(self,
                                     app_name: str,
                                     ip_address: str,
                                     script_name: str) -> bool:

        self.logger.info("entered '{}'".format(app_name))
        script_file_path = Const.get_health_check_file(app_name=app_name, script_name=script_name)
        command = ['{script_file_path} {ip_address}'.format(
            script_file_path=script_file_path,
            ip_address=ip_address)]

        app_folder = Const.get_app_folder(app_name=app_name)
        if not self._file_service.exists(app_folder):
            self._file_service.create_folders(app_folder)

        with open(Const.get_app_log_file(app_name=app_name), 'ab', 0) as output_file:
            timestamp = Utils.get_timestamp()
            output_file.write('health-check started: {0} with command: {1}\n'.format(timestamp, command).encode())

            with subprocess.Popen(command,
                                  stdout=output_file,
                                  stderr=output_file,
                                  shell=True,
                                  preexec_fn=os.setsid) as p:
                try:
                    p.communicate(timeout=self.timeout)
                except subprocess.TimeoutExpired:
                    output_file.write('health-check: timeout after {} seconds\n'.format(self.timeout).encode())
                finally:
                    if p.returncode:
                        output_file.write('health-check: error, script exit with "{}"\n'.format(p.returncode).encode())
                        self._kill_process(p)
                    else:
                        if p.returncode == 0:
                            output_file.write('health-check: done\n'.encode())
                            return True
                        else:
                            self._kill_process(p)
                            return False

    @staticmethod
    def _kill_process(p: subprocess.Popen):
        try:
            os.killpg(os.getpgid(p.pid), signal.SIGTERM)
        except ProcessLookupError:
            print('process already killed')



from sidecar.app_instance_identifier import AppInstanceIdentifier
from sidecar.app_instance_identifier_creator import IAppInstanceIdentifierCreator
from sidecar.kub_api_pod_reader import KubApiPodReader
from sidecar.kub_api_service import KubApiService


class KubAppInstanceIdentifierCreator(IAppInstanceIdentifierCreator):
    def __init__(self, kub_api_service: KubApiService) -> None:
        super().__init__()
        self._kub_api_service = kub_api_service

    def create(self, app_name: str, ip_address: str, additional_data: {}) -> AppInstanceIdentifier:
        # TODO: change ip address to pod name/id (or specify it in additional_data) and use it to find the pod because:
        #       a. consistent with aws
        #       b. don't want to have to handle the situation where ip is not available on the pod (maybe can happen
        #          in some pod phases, like 'pending'), which will cause the pod not to be found
        pod_json = self._kub_api_service.try_get_pod_json(ip_address=ip_address)
        if not pod_json:
            raise Exception("pod with address '{ADDRESS}' not found".format(ADDRESS=ip_address))
        # container id is enough as infra identifier because it is supposed to be unique (across space and time),
        # meaning that even if there are multiple instances of the same app on different pods -
        # they should have different container ids
        container_id = KubApiPodReader.safely_get_container_id_for_app(app_name=app_name, pod=pod_json)
        if not container_id:
            raise Exception("container id for app '{APP}' not found on pod with address '{ADDRESS}'"
                            .format(APP=app_name, ADDRESS=ip_address))
        return AppInstanceIdentifier(name=app_name, infra_id=container_id)
